#! /usr/bin/env python3

import logging
import re
import argparse
import sys
from math import log2
from collections import deque
from time import perf_counter as timer


class ParseError(Exception):
    pass


class BDD:
    table = {}                      # Unique Table
    inv_table = {}                  # Inverse of the Unique Table
    index = 2                       # Next index to be used
    v_table = {0: set(), 1: set()}  # Set of variables in a BDD
    computed = {}                   # Memo table for Binary operations
    neg_table = {}                  # Memo table for NOT operation

    def __init__(self, root):
        self.root = root            # Index of the BDD in the unique table

    @staticmethod
    def mk(x, low, high):
        """Find-or-create node in the unique table (returns index)"""
        if low == high:
            return low
        try:
            return BDD.inv_table[(x, low, high)]
        except KeyError:
            index = BDD.index
            BDD.table[index] = (x, low, high)
            BDD.v_table[index] = BDD.v_table[low].union(BDD.v_table[high])
            BDD.v_table[index].add(x)
            BDD.inv_table[(x, low, high)] = index
            BDD.index += 1
        return index

    @classmethod
    def from_unit(cls, literal):
        pos_index = cls.mk(abs(literal), 0, 1)
        neg_index = cls.mk(abs(literal), 1, 0)

        cls.neg_table[pos_index] = neg_index
        cls.neg_table[neg_index] = pos_index

        if literal == abs(literal):
            f = cls(pos_index)
        else:
            f = cls(neg_index)

        BDD.v_table[neg_index] = {abs(literal)}
        BDD.v_table[pos_index] = {abs(literal)}
        return f

    @classmethod
    def from_cube(cls, literals):
        literals.sort(key=lambda x: abs(x))
        f = cls(1)
        for lit in literals:
            f = apply(f, cls.from_unit(lit), 'and')
        return f

    @classmethod
    def from_clause(cls, literals):
        literals.sort(key=lambda x: abs(x))
        f = cls(0)
        for lit in literals:
            f = apply(f, cls.from_unit(lit), 'or')
        return f

    @classmethod
    def from_xor(cls, literals):
        literals.sort(key=lambda x: abs(x))
        f = cls(0)
        for lit in literals:
            f = apply(f, cls.from_unit(lit), 'xor')
        return f

    @staticmethod
    def lo(node):
        return BDD.table[node][1]

    @staticmethod
    def hi(node):
        return BDD.table[node][2]

    @staticmethod
    def variable(node):
        return BDD.table[node][0]

    @staticmethod
    def negated(node):
        return BDD.neg_table[node]

    def is_bot(self):
        return self.root == 0

    def is_top(self):
        return self.root == 1

    def cube_literals(self):
        """Returns list of literals if the BDD is cubic"""
        literals = []
        node = self.root
        while node != 0 and node != 1:
            if BDD.lo(node) == 0:
                literals.append(BDD.variable(node))
                node = BDD.hi(node)
            elif BDD.hi(node) == 0:
                literals.append(-BDD.variable(node))
                node = BDD.lo(node)
            else:
                return []
        return literals

    def units(self):
        literals = []
        for v in self.variables():
            positive = BDD.from_unit(v)
            if cofactor(self, positive).is_bot():
                literals.append(-v)
            else:
                negative = BDD.from_unit(-v)
                if cofactor(self, negative).is_bot():
                    literals.append(v)
        return literals

    def variables(self):
        return BDD.v_table[self.root]

    @staticmethod
    def explore(node, path=None):
        """Used by the paths function below"""

        if path is None:
            path = []

        if node == 1:
            yield path
        elif BDD.lo(node) == 0 or BDD.hi(node) == 0:
            if BDD.lo(node) == 0:
                lit = BDD.variable(node)
                next_node = BDD.hi(node)
            else:
                lit = -BDD.variable(node)
                next_node = BDD.lo(node)
            path.append(lit)
            yield from BDD.explore(next_node, path)
        else:
            lo_path = path + [-BDD.variable(node)]
            yield from BDD.explore(BDD.lo(node), lo_path)
            hi_path = path + [BDD.variable(node)]
            yield from BDD.explore(BDD.hi(node), hi_path)

    def paths(self):
        """Returns all paths to the True terminal in the BDD"""
        return BDD.explore(self.root)


def apply(left, right, op):
    # Standard BDD apply function
    def t_and(t1, t2):
        if t1 == 1 and t2 == 1:
            return 1
        else:
            return 0

    def t_or(t1, t2):
        if t1 == 1 or t2 == 1:
            return 1
        else:
            return 0

    def t_xor(t1, t2):
        if t1 == 1 and t2 == 1:
            return 0
        elif t1 == 0 and t2 == 0:
            return 0
        else:
            return 1

    def t_not(t1):
        if t1 == 0:
            return 1
        else:
            return 0

    if op == 'and':
        t_op = t_and
    elif op == 'or':
        t_op = t_or
    elif op == 'xor':
        t_op = t_xor

    def app(node1, node2):
        if (node1, node2, op) in BDD.computed:
            u = BDD.computed[(node1, node2, op)]
            return u, BDD.neg_table[u]
        elif node1 in [0, 1] and node2 in [0, 1]:
            u = t_op(node1, node2)
            neg_u = t_not(u)
            return u, neg_u
        elif node1 in [0, 1]:
            var = BDD.variable(node2)
            lo, neg_lo = app(node1, BDD.lo(node2))
            hi, neg_hi = app(node1, BDD.hi(node2))
            u = BDD.mk(var, lo, hi)
            neg_u = BDD.mk(var, neg_lo, neg_hi)
        elif node2 in [0, 1]:
            var = BDD.variable(node1)
            lo, neg_lo = app(BDD.lo(node1), node2)
            hi, neg_hi = app(BDD.hi(node1), node2)
            u = BDD.mk(var, lo, hi)
            neg_u = BDD.mk(var, neg_lo, neg_hi)
        elif BDD.variable(node1) == BDD.variable(node2):
            var = BDD.variable(node1)
            lo, neg_lo = app(BDD.lo(node1), BDD.lo(node2))
            hi, neg_hi = app(BDD.hi(node1), BDD.hi(node2))
            u = BDD.mk(var, lo, hi)
            neg_u = BDD.mk(var, neg_lo, neg_hi)
        elif BDD.variable(node1) < BDD.variable(node2):
            var = BDD.variable(node1)
            lo, neg_lo = app(BDD.lo(node1), node2)
            hi, neg_hi = app(BDD.hi(node1), node2)
            u = BDD.mk(var, lo, hi)
            neg_u = BDD.mk(var, neg_lo, neg_hi)
        else:
            var = BDD.variable(node2)
            lo, neg_lo = app(node1, BDD.lo(node2))
            hi, neg_hi = app(node1, BDD.hi(node2))
            u = BDD.mk(var, lo, hi)
            neg_u = BDD.mk(var, neg_lo, neg_hi)
        BDD.computed[(node1, node2, op)] = u
        BDD.neg_table[u] = neg_u
        return u, neg_u

    return BDD(app(left.root, right.root)[0])


def cofactor(bdd, constraint):
    def reduce_1(node, u):
        assert u > 0
        if BDD.variable(node) == u:
            return BDD.hi(node)
        else:
            return node

    def reduce_0(node, u):
        assert u > 0
        if BDD.variable(node) == u:
            return BDD.lo(node)
        else:
            return node

    def sim(node1, node2):
        if (node1, node2, 'cof') in BDD.computed:
            u = BDD.computed[(node1, node2, 'cof')]
            return u, BDD.neg_table[u]
        if node1 == 0 or (node2 in [0, 1] and node2 == 0):
            return 0, 1
        if node1 == 1:
            return 1, 0
        if node2 in [0, 1] and node2 == 1:
            return node1, BDD.neg_table[node1]
        var = min(BDD.variable(node1), BDD.variable(node2))
        if reduce_0(node2, var) == 0:
            u, neg_u = sim(reduce_1(node1, var), reduce_1(node2, var))
        elif reduce_1(node2, var) == 0:
            u, neg_u = sim(reduce_0(node1, var), reduce_0(node2, var))
        else:
            lo, neg_lo = sim(reduce_0(node1, var), reduce_0(node2, var))
            hi, neg_hi = sim(reduce_1(node1, var), reduce_1(node2, var))
            u = BDD.mk(var, lo, hi)
            neg_u = BDD.mk(var, neg_lo, neg_hi)
        BDD.computed[(node1, node2, 'cof')] = u
        BDD.neg_table[u] = neg_u
        return u, neg_u

    return BDD(sim(bdd.root, constraint.root)[0])


class Solver:

    def __init__(self):
        self.clauses = []           # Accumulated clauses
        self.inv_clauses = {}       # str(clause) -> its index in self.clauses
        self.clauses_by_lit = {}    # literal -> list of clause (indices) in which it occurs

        self.diagrams = {}          # Accumulated BDDs, key is just index in unique table
        self.inv_diagrams = {}      # str(bdd) -> index in self.diagrams
        self.diagrams_by_var = {}   # variable -> list of BDDs (indices) variable occurs in
        self.alpha_diagrams = {}    # index -> BDD at index reduced by current alpha

        self.assignments = {}       # Current assignment (by literal)
        self.prop_queue = deque()   # Literals assigned since last propagation
        self.alpha_list = []        # Literals assigned at the alpha-level
        self.verified = False       # Set to True when the empty clause is derivable

        self.counts = {'path': 0, 'rup-0': 0, 'rup': 0}
        self.size = 0

    @classmethod
    def parse_formula(cls, formula_file):
        solver = cls()

        # Regular expressions for different lines in the CNF
        info_line = re.compile("p\s*cnf\s*(\d*)\s*(\d*)")
        comment_line = re.compile("c.*")
        clause_line = re.compile("((?:-?\d+\s+)*)0$")

        clauses_seen = 0
        seen_info = False

        started = False

        with open(formula_file) as formula:
            for line_no, line in enumerate(formula):
                line = line.strip()
                if not line or comment_line.match(line):
                    continue

                started = True

                if not seen_info:
                    info = info_line.match(line)
                    if info:
                        clause_count = int(info.group(2))
                        seen_info = True
                        logging.info(f"Parsing formula with {info.group(1)} "
                                     f"variables and {clause_count} clauses")
                        continue

                clause_match = clause_line.match(line)
                if not clause_match:
                    raise ParseError(f"line {line_no + 1}")

                try:
                    literals = [int(lit) for lit in clause_match.group(1).split()]
                except (IndexError, ValueError):
                    raise ParseError(f"line {line_no + 1}")

                if 0 in literals:
                    raise ParseError(f"line {line_no + 1}")

                if not literals:
                    logging.warning("* Empty clause in input formula *")
                    solver.verified = True
                    return solver

                if len(literals) > 1:
                    solver.add_clause(literals)
                elif not solver.assign(literals[0]):
                    logging.warning("* Conflicting unit clauses in formula *")
                    solver.verified = True
                    return solver

                clauses_seen += 1

        if not started:
            raise ParseError("empty formula")

        if clauses_seen != clause_count:
            raise ParseError("unexpected number of clauses")

        return solver

    def add_clause(self, clause):
        clause.sort()
        if len(clause) > 1:
            self.clauses.append(clause)
            self.inv_clauses[str(clause)] = len(self.clauses) - 1
            self.size += len(clause) + 2
            for literal in clause:
                if literal not in self.clauses_by_lit:
                    self.clauses_by_lit[literal] = [clause]
                else:
                    self.clauses_by_lit[literal].append(clause)
        elif len(clause) == 1:
            if not self.assign(clause[0]):
                # New unit conflicts with fixed assignment
                self.verified = True
                return False
        return True

    def add_diagram(self, bdd, id_string, size=0):
        assert not bdd.is_top() and not bdd.is_bot()

        if BDD.neg_table[bdd.root] in self.diagrams:
            self.verified = True
            logging.info("Negated BDD already present")
            return False

        cube = bdd.cube_literals()
        if cube:
            logging.debug("BDD to add is cubic")
            # cofactor existing diagrams by bdd
            for literal in cube:
                if not self.assign(literal):
                    self.verified = True
                    return False
            if not self.reduce_by_bdd(bdd):
                logging.info(f"Existing BDD is falsified after adding cube")
                self.verified = True
                return False
            return True

        self.diagrams[bdd.root] = bdd
        self.inv_diagrams[id_string] = bdd.root
        self.size += size

        for variable in bdd.variables():
            if variable in self.diagrams_by_var:
                self.diagrams_by_var[variable].appendleft(bdd.root)
            else:
                self.diagrams_by_var[variable] = deque([bdd.root])
        return True

    def assign(self, lit, level=None):
        if lit in self.assignments:
            return self.assignments[lit]
        self.assignments[lit] = True
        self.assignments[-lit] = False
        self.prop_queue.append(lit)
        if level == 'alpha':
            self.alpha_list.append(lit)
        return True

    def clear_alpha_assign(self):
        self.prop_queue = deque()
        while self.alpha_list:
            self.unassign(self.alpha_list.pop())
        self.alpha_diagrams = {}

    def unassign(self, lit):
        assert lit in self.assignments and -lit in self.assignments
        del self.assignments[lit]
        del self.assignments[-lit]

    def delete_clause(self, clause):
        clause.sort()
        clause_string = str(clause)
        if clause_string in self.inv_clauses:
            index = self.inv_clauses[clause_string]
            self.clauses[index] = ['T']
            self.size -= len(clause) + 2
            return True
        else:
            logging.info("* Clause to delete not found in clause set *")
            return False

    def delete_diagram(self, diagram_string, size=0):
        if diagram_string in self.inv_diagrams:
            key = self.inv_diagrams[diagram_string]
            assert key in self.diagrams
            self.diagrams[key] = BDD(1)
            self.size -= size
        else:
            logging.info("* Diagram to delete not found in accumulated diagrams")
            return False

    def get_diagram_at_level(self, index, level=None):
        if level is None:
            return self.diagrams[index]
        else:
            assert level == 'alpha'
            if index in self.alpha_diagrams:
                return self.alpha_diagrams[index]
            else:
                return self.diagrams[index]

    def set_diagram_at_level(self, index, diagram, level=None):
        if level is None:
            self.diagrams[index] = diagram
        else:
            assert level == 'alpha'
            self.alpha_diagrams[index] = diagram

    def bdds_with_variables(self, variables):
        """Returns list of BDD indices in which some v in 'variables' occurs"""
        diagrams = set()
        list_diags = deque([])
        for v in variables:
            if v not in self.diagrams_by_var:
                continue
            for index in self.diagrams_by_var[v]:
                if index in diagrams:
                    continue
                if self.diagrams[index].is_top():
                    continue
                diagrams.add(index)
                list_diags.append(index)
        return list_diags

    def propagate(self, level=None):
        # Propagate over the clause set
        record = []  # remember propagated literals
        while self.prop_queue:
            falsified = -self.prop_queue.popleft()
            record.append(-falsified)
            if falsified not in self.clauses_by_lit:
                continue
            for clause in self.clauses_by_lit[falsified]:
                if clause == ['T']:
                    continue
                unassigned = []
                satisfied = False
                for lit in clause:
                    if lit not in self.assignments:
                        unassigned.append(lit)
                        if len(unassigned) > 1:
                            break
                    elif self.assignments[lit]:
                        satisfied = True
                        break
                if satisfied:
                    continue
                if len(unassigned) == 1:
                    self.assign(unassigned[0], level=level)
                elif len(unassigned) == 0:
                    return False

        # Now propagate over the BDDs
        if self.diagrams:
            if not self.bdd_propagate(level=level, head_start=record):
                return False
            elif self.prop_queue:
                return self.propagate(level=level)

        return True

    def bdd_propagate(self, level=None, head_start=None):
        while self.prop_queue or head_start:
            if head_start:
                prop_cube = BDD.from_cube(head_start)
                head_start = []
            else:
                prop_cube = BDD.from_cube(list(self.prop_queue))
                self.prop_queue = deque()
            if not self.reduce_by_bdd(prop_cube, level=level):
                logging.debug("Propagation over BDDs produces conflict")
                return False
        return True

    def reduce_by_bdd(self, bdd, level=None):
        assert not bdd.is_top() and not bdd.is_bot()
        cube = bdd.cube_literals()
        if not cube:
            cube = bdd.units()
            
        for literal in cube:
            if not self.assign(literal, level=level):
                return False

        if not self.diagrams:
            return True
        
        negations_seen = set()
        for index in self.bdds_with_variables(bdd.variables()):
            if self.diagrams[index].is_top():
                continue
            diagram = self.get_diagram_at_level(index, level=level)
            diagram_reduced = cofactor(diagram, bdd)
            if diagram_reduced.is_bot():
                return False
            if diagram_reduced.root in negations_seen:
                return False
            if BDD.negated(diagram_reduced.root) in self.diagrams:
                return False
            negations_seen.add(BDD.negated(diagram_reduced.root))
            units = diagram_reduced.units()
            if units:
                for literal in units:
                    if not self.assign(literal, level=level):
                        return False
                continue
            self.set_diagram_at_level(index, diagram_reduced, level=level)
        return True

    def path_check_alpha(self, alpha):
        # We increase the log-bound by a constant factor to account for how
        # Lingeling extracts some XOR constraints from CNF formulas
        if 2 * len(alpha.variables()) - 1 > 2 * log2(self.size):
            logging.info("*** Size of BDD exceeds size for path-wise RUP check ***")
            return False

        # Need conflict on all paths
        for path in alpha.paths():
            for lit in path:
                if not self.assign(lit, level='alpha'):
                    self.clear_alpha_assign()
                    continue
            if self.propagate(level='alpha'):
                self.clear_alpha_assign()
                return False
            self.clear_alpha_assign()

        self.counts["path"] += 1
        return True

    def check_bdd(self, bdd, is_xor):
        """Returns True if the bdd is RUP wrt F and Phi"""
        alpha = BDD(BDD.negated(bdd.root))

        assert not self.alpha_diagrams

        if not self.reduce_by_bdd(alpha, level='alpha'):
            # Then alpha diagrams already contradictory
            self.clear_alpha_assign()
            self.counts["rup-0"] += 1
            return True, "rup-0"

        if not self.propagate(level='alpha'):
            # Then propagating alpha diagrams produces conflict
            self.clear_alpha_assign()
            self.counts["rup"] += 1
            return True, "rup"

        self.clear_alpha_assign()

        # Check for path-wise addition (if not a clause)
        if not is_xor:
            return False, ""
        path_check = self.path_check_alpha(alpha)
        if path_check:
            return True, "path"
        else:
            return False, ""


def parse_proof_line(line):
    """
    Parses a single line from a proof file. If the line starts with a 'd', we
    interpret it as a deletion instruction, otherwise as an addition. If the
    constraint starts with an 'x' we interpret it as an XOR clause, otherwise
    assume it is a clause. Note: to delete an XOR, the 'd' must precede the 'x'
    """

    instruction = re.compile("(d?)\s*(x?)\s*((?:-?\d+\s+)*)0\s*")
    line_match = instruction.match(line.rstrip('\n'))

    if not line_match:
        raise ParseError("malformed proof instruction")

    delete = line_match.group(1)
    xor_clause = line_match.group(2)
    literals = [int(n) for n in line_match.group(3).split()]
    try:
        if line_match.group(4):
            raise ParseError("malformed proof instruction")
    except IndexError:
        pass

    if 0 in literals:
        raise ParseError("malformed proof instruction")

    return literals, xor_clause, delete


def proof_check(solver, proof_file, persist):
    comment_line = re.compile('c.*')

    # Count the number of failed additions
    failures = 0

    line_offset = 0

    started = False

    with open(proof_file) as proof:
        for line_no, line_in in enumerate(proof):
            failed_addition = False
            line = line_in.strip()

            if not line or comment_line.match(line):
                line_offset += 1
                continue

            started = True

            try:
                clause, xor, delete = parse_proof_line(line)
            except ParseError as e:
                raise ParseError(f"{e} (line {line_no - line_offset + 1})")

            if delete:
                clause.sort()
                if xor:
                    del_str = "XOR " + str(clause)
                    solver.delete_diagram(del_str, size=2*len(clause)+1)
                    logging.info(f"{line_no - line_offset + 1}: deleted XOR {clause}")
                else:
                    solver.delete_clause(clause)
                    logging.info(f"{line_no - line_offset + 1}: deleted {clause}")
                continue

            # We check empty clause addition afterwards anyway
            if not clause:
                break

            if xor:
                add_bdd = BDD.from_xor(clause)
                add_str = "XOR " + str(clause)
            else:
                add_bdd = BDD.from_clause(clause)
                add_str = str(clause)

            # Try to add the constraint
            result, message = solver.check_bdd(add_bdd,is_xor=xor)

            if not result:
                logging.critical(f"*** Addition failed: {add_str} ***")
                failed_addition = True
                failures += 1
                if not persist:
                    break
            else:
                if message == "rup-0" or message == "rup":
                    logging.info(f"{line_no - line_offset + 1}: added {add_str}"
                                 " by RUP_BDD")
                elif message == "path":
                    logging.info(f"{line_no - line_offset + 1}: added {add_str}"
                                 " by RUP_path")

            # If addition succeeds, add the diagram
            if not failed_addition:
                if xor:
                    if not solver.add_diagram(add_bdd, id_string=add_str, size=2*len(clause)+1):
                        logging.info("XOR addition leads to conflict")
                        solver.verified = True
                        break
                else:
                    if not solver.add_clause(clause):
                        logging.info("Clause addition leads to conflict")
                        solver.verified = True
                        break
                    if len(clause) == 1:
                        logging.debug("Propagating after adding unit clause")
                        if not solver.propagate():
                            logging.debug("Propagation leads to conflict")
                            solver.verified = True
                            break
    if not started:
        raise ParseError("empty proof")
    if failed_addition and not persist:
        return False

    if failures == 0:
        logging.debug("All addition steps succeeded")
    else:
        logging.critical(f"* {failures} failed addition steps *")

    if solver.verified:
        logging.debug("Empty clause implied after proof")
        return solver.verified

    logging.debug("Checking for empty clause addition after proof")
    if not solver.propagate():
        logging.debug("Empty clause implied after proof")
        solver.verified = True
    else:
        logging.critical("*** False not implied after proof ***")
        solver.verified = False

    return solver.verified


def main():
    argparser = argparse.ArgumentParser()

    argparser.add_argument("formula", metavar="<formula>", action="store",
                           help="path to CNF formula")

    argparser.add_argument("proof", metavar="<proof>", action="store",
                           help="path to proof file")

    argparser.add_argument("-v", "--verbose", action="count", default=0,
                           help="increase output verbosity [default: quiet]")

    argparser.add_argument("-p", "--persist", action="store_true",
                           default=False, help="ignore non-critical addition"
                                               " failures [default: off]")

    args = argparser.parse_args()

    if args.verbose == 0:
        logging.basicConfig(format="c %(message)s", level=logging.WARNING,
                            stream=sys.stdout)
    elif args.verbose == 1:
        logging.basicConfig(format="c %(message)s", level=logging.INFO,
                            stream=sys.stdout)
    else:
        logging.basicConfig(format="c %(message)s", level=logging.DEBUG,
                            stream=sys.stdout)

    start = timer()

    try:
        solver = Solver.parse_formula(args.formula)
    except FileNotFoundError:
        sys.exit("*** Formula file not found ***")
    except ParseError as e:
        sys.exit(f"*** Error parsing formula: {e} ***")

    logging.info("Finished parsing")

    if solver.verified:
        logging.critical("*** Input formula trivially unsatisfiable ***")
        sys.exit(0)

    if solver.assignments:
        logging.debug("Propagating unit clauses in input formula")
        if not solver.propagate():
            logging.critical("*** Propagation on input formula produces"
                             " conflict ***")
            solver.verified = True

    if not solver.verified:
        logging.info("Starting proof checking")
        if args.persist:
            logging.critical("--persist: Ignoring (non-False) addition"
                             " failures")
        try:
            proof_check(solver, args.proof, args.persist)
        except FileNotFoundError:
            sys.exit("*** Proof file not found ***")
        except ParseError as e:
            sys.exit(f"*** Error parsing proof: {e} ***")
        logging.info("Finished proof checking")

    end = timer()

    if solver.verified:
        print("s VERIFIED")
        logging.debug(f"RUP_path additions: {solver.counts['path']}")
        logging.debug(f"RUP_bdd  additions: "
                      f"{solver.counts['rup-0'] + solver.counts['rup']}")
    else:
        print("s NOT VERIFIED")

    time = end - start

    logging.critical(f"Elapsed time (sec): {time:.2f}")
    sys.exit(0)


if __name__ == "__main__":
    main()
